package org.apache.bcel.verifier.structurals;

package .tmp;

import java.util.HashSet;
import java.util.Set;
import org.apache.bcel.generic.ASTORE;
import org.apache.bcel.generic.IndexedInstruction;
import org.apache.bcel.generic.Instruction;
import org.apache.bcel.generic.InstructionHandle;
import org.apache.bcel.generic.JsrInstruction;
import org.apache.bcel.generic.LocalVariableInstruction;
import org.apache.bcel.generic.RET;
import org.apache.bcel.verifier.exc.AssertionViolatedException;
import org.apache.bcel.verifier.exc.StructuralCodeConstraintException;
import org.apache.bcel.verifier.structurals.Subroutine;
import org.apache.bcel.verifier.structurals.Subroutines;

class SubroutineImpl implements Subroutine {
  private static final int UNSET = -1;
  
  private final SubroutineImpl[] EMPTY_ARRAY = new SubroutineImpl[0];
  
  private int localVariable = -1;
  
  private final Set<InstructionHandle> instructions = new HashSet<>();
  
  private final Set<InstructionHandle> theJSRs = new HashSet<>();
  
  private InstructionHandle theRET;
  
  private void _getRecursivelyAccessedLocalsIndicesHelper(Set<Integer> set, Subroutine[] subs) {
    for (Subroutine sub : subs) {
      int[] lvs = sub.getAccessedLocalsIndices();
      for (int lv : lvs)
        set.add(Integer.valueOf(lv)); 
      if ((sub.subSubs()).length != 0)
        _getRecursivelyAccessedLocalsIndicesHelper(set, sub.subSubs()); 
    } 
  }
  
  public void addEnteringJsrInstruction(InstructionHandle jsrInst) {
    if (jsrInst == null || !(jsrInst.getInstruction() instanceof JsrInstruction))
      throw new AssertionViolatedException("Expecting JsrInstruction InstructionHandle."); 
    if (this.localVariable == -1)
      throw new AssertionViolatedException("Set the localVariable first!"); 
    if (this.localVariable != ((ASTORE)((JsrInstruction)jsrInst.getInstruction()).getTarget().getInstruction()).getIndex())
      throw new AssertionViolatedException("Setting a wrong JsrInstruction."); 
    this.theJSRs.add(jsrInst);
  }
  
  void addInstruction(InstructionHandle ih) {
    if (this.theRET != null)
      throw new AssertionViolatedException("All instructions must have been added before invoking setLeavingRET()."); 
    this.instructions.add(ih);
  }
  
  public boolean contains(InstructionHandle inst) {
    return this.instructions.contains(inst);
  }
  
  public int[] getAccessedLocalsIndices() {
    Set<Integer> acc = new HashSet<>();
    if (this.theRET == null && this != Subroutines.this.getTopLevel())
      throw new AssertionViolatedException("This subroutine object must be built up completely before calculating accessed locals."); 
    for (InstructionHandle ih : this.instructions) {
      if (ih.getInstruction() instanceof LocalVariableInstruction || ih.getInstruction() instanceof RET) {
        int idx = ((IndexedInstruction)ih.getInstruction()).getIndex();
        acc.add(Integer.valueOf(idx));
        try {
          if (ih.getInstruction() instanceof LocalVariableInstruction) {
            int s = ((LocalVariableInstruction)ih.getInstruction()).getType(null).getSize();
            if (s == 2)
              acc.add(Integer.valueOf(idx + 1)); 
          } 
        } catch (RuntimeException re) {
          throw new AssertionViolatedException("BCEL did not like NULL as a ConstantPoolGen object.", re);
        } 
      } 
    } 
    int[] ret = new int[acc.size()];
    int j = -1;
    for (Integer accessedLocal : acc) {
      j++;
      ret[j] = accessedLocal.intValue();
    } 
    return ret;
  }
  
  public InstructionHandle[] getEnteringJsrInstructions() {
    if (this == Subroutines.this.getTopLevel())
      throw new AssertionViolatedException("getLeavingRET() called on top level pseudo-subroutine."); 
    return this.theJSRs.<InstructionHandle>toArray(InstructionHandle.EMPTY_ARRAY);
  }
  
  public InstructionHandle[] getInstructions() {
    return this.instructions.<InstructionHandle>toArray(InstructionHandle.EMPTY_ARRAY);
  }
  
  public InstructionHandle getLeavingRET() {
    if (this == Subroutines.this.getTopLevel())
      throw new AssertionViolatedException("getLeavingRET() called on top level pseudo-subroutine."); 
    return this.theRET;
  }
  
  public int[] getRecursivelyAccessedLocalsIndices() {
    Set<Integer> s = new HashSet<>();
    int[] lvs = getAccessedLocalsIndices();
    for (int lv : lvs)
      s.add(Integer.valueOf(lv)); 
    _getRecursivelyAccessedLocalsIndicesHelper(s, subSubs());
    int[] ret = new int[s.size()];
    int j = -1;
    for (Integer index : s) {
      j++;
      ret[j] = index.intValue();
    } 
    return ret;
  }
  
  void setLeavingRET() {
    if (this.localVariable == -1)
      throw new AssertionViolatedException("setLeavingRET() called for top-level 'subroutine' or forgot to set local variable first."); 
    InstructionHandle ret = null;
    for (InstructionHandle actual : this.instructions) {
      if (actual.getInstruction() instanceof RET) {
        if (ret != null)
          throw new StructuralCodeConstraintException("Subroutine with more then one RET detected: '" + ret + "' and '" + actual + "'."); 
        ret = actual;
      } 
    } 
    if (ret == null)
      throw new StructuralCodeConstraintException("Subroutine without a RET detected."); 
    if (((RET)ret.getInstruction()).getIndex() != this.localVariable)
      throw new StructuralCodeConstraintException("Subroutine uses '" + ret + "' which does not match the correct local variable '" + this.localVariable + "'."); 
    this.theRET = ret;
  }
  
  void setLocalVariable(int i) {
    if (this.localVariable != -1)
      throw new AssertionViolatedException("localVariable set twice."); 
    this.localVariable = i;
  }
  
  public Subroutine[] subSubs() {
    Set<Subroutine> h = new HashSet<>();
    for (InstructionHandle ih : this.instructions) {
      Instruction inst = ih.getInstruction();
      if (inst instanceof JsrInstruction) {
        InstructionHandle targ = ((JsrInstruction)inst).getTarget();
        h.add(Subroutines.this.getSubroutine(targ));
      } 
    } 
    return h.<Subroutine>toArray((Subroutine[])this.EMPTY_ARRAY);
  }
  
  public String toString() {
    StringBuilder ret = new StringBuilder();
    ret.append("Subroutine: Local variable is '").append(this.localVariable);
    ret.append("', JSRs are '").append(this.theJSRs);
    ret.append("', RET is '").append(this.theRET);
    ret.append("', Instructions: '").append(this.instructions).append("'.");
    ret.append(" Accessed local variable slots: '");
    int[] alv = getAccessedLocalsIndices();
    for (int element : alv) {
      ret.append(element);
      ret.append(" ");
    } 
    ret.append("'.");
    ret.append(" Recursively (via subsub...routines) accessed local variable slots: '");
    alv = getRecursivelyAccessedLocalsIndices();
    for (int element : alv) {
      ret.append(element);
      ret.append(" ");
    } 
    ret.append("'.");
    return ret.toString();
  }
}
